package internetBanking;

/**
 * @author sergedemeyer - September 2015 An account is a dataholder for a
 *         balance (the amount of money available) and is identified by means of
 *         an accountNr (a unique identifier)
 */

public class Account extends DesignByContract {
	private int _accountNr = 0, _balance = 0;
	private int _transactionId = 0, _workingBalance = 0;

	public Account(int accountNr) {
		this._accountNr = accountNr;
	}

	/**
	 * @return the _accountNr
	 */
	public int get_accountNr() {
		return _accountNr;
	}

	/**
	 * @return the _balance, used as a peeking method for checking
	 *         pre-conditions within contracts
	 */
	protected synchronized int peek_balance() {
		return _balance;
	}

	/**
	 * @param transaction
	 *            identifies the transaction accessing the balance
	 * @return the _balance
	 */
	public synchronized int get_balance(int transaction) {
		this.require(this.isLockedBy(transaction), "Lock is required to operate on an account.");
		return _workingBalance;
	}

	/**
	 * @param balance
	 *            the amount to be added to _balance
	 * @param transaction
	 *            identifies the transaction accessing the balance
	 */
	public synchronized void inc_balance(int balance, int transaction) {
		this.require(this.isLockedBy(transaction), "Lock is required to operate on an account.");
		this._workingBalance = _workingBalance + balance;
	}

	/**
	 * Lock this account, ensure that only transactions holding the appropriate
	 * transaction may access the account
	 * 
	 * @param transaction
	 *            identifies the transaction accessing the balance
	 */
	public synchronized void lock(int transaction) {
		this.require(this.notLocked(), "No other transaction may hold a lock.");
		this._transactionId = transaction;
		this._workingBalance = this._balance;
		this.ensure(this.isLockedBy(transaction), "Lock must be held by current transaction.");
	}

	/**
	 * Commit the pending state-changes on this account, if they were performed
	 * by the transactions holding the appropriate transaction
	 * 
	 * @param transaction
	 *            identifies the transaction accessing the balance
	 */
	public synchronized void commit(int transaction) {
		this.require(this.isLockedBy(transaction), "Lock is required to commit transaction.");
		this._transactionId = 0;
		this._balance = this._workingBalance;
		this.ensure(this.notLocked(), "Lock should be released after committing the transaction.");
	}

	/**
	 * Commit the pending state-changes on this account, if they were performed
	 * by the transactions holding the appropriate transaction
	 * 
	 * @param transaction
	 *            identifies the transaction accessing the balance
	 */
	public synchronized void abort(int transaction) {
		this.require(this.isLockedBy(transaction), "Lock is required to abort transaction.");
		this._transactionId = 0;
		this._workingBalance = 0;		
		this.ensure(this.notLocked(), "Lock should be released after aborting the transaction.");
	}

	/**
	 * Answer whether no transaction is currently locking the account
	 */
	public synchronized boolean notLocked() {
		return this._transactionId == 0;
	}

	/**
	 * Answer whether this account is currently locked by the given transaction
	 * 
	 * @param transaction
	 *            identifies the transaction accessing the balance
	 */
	public synchronized boolean isLockedBy(int transaction) {
		return this._transactionId != 0;
	}

}
