package internetBanking;

import java.util.ArrayList;
import java.util.Iterator;

public class Bank extends DesignByContract {
	/**
	 * Class Bank manages the associations between customer and accounts. The
	 * class provides the public interface used to deposit money on accounts ;
	 * see the balance of an account and transfer money between accounts
	 */

	/**
	 * @author sergedemeyer - September 2015
	 *
	 */

	/**
	 * _accounts maintains an list of all account numbers. Account numbers range
	 * from 1 to size of _accounts. Access should always be via index - 1
	 */
	private ArrayList<Account> _accounts = new ArrayList<Account>(5);
	private static int transactionCounter = 0;

	/**
	 * @return answer whether a given customer is a valid customer for this bank
	 */
	private static int nextTransactionId() {
		return ++transactionCounter;
	}

	/**
	 * @return answer whether a given customer is a valid customer for this bank
	 */
	public boolean valid_customer(Customer customer) {
		return true;
	}

	/**
	 * @return return the account number for a newly created account
	 */
	public int createAccountForCustomer(Customer customer) {
		this.require(this.valid_customer(customer), "Invalid customer");
		int accountNumber = _accounts.size();
		Account account = new Account(accountNumber);
		_accounts.add(account);
		this.ensure(this.customerMayAccess(customer, accountNumber + 1), "Customer cannot access acount");
		return accountNumber + 1;
	}

	/**
	 * @return answer whether a given customer may access the account associated
	 *         with the given account number
	 */
	public boolean customerMayAccess(Customer customer, int accountNumber) {
		return (accountNumber > 0) && (accountNumber <= _accounts.size());
	}

	/**
	 * @return return the account number for a newly created account
	 */
	public int seeBalance(Customer customer, int accountNumber) {
		int tmpBalance = 0;
		this.invariant(this.checkSumOfAccounts(), "Inconsistencies between account balances -- checksum incorrect.");
		this.require(this.valid_customer(customer), "Invalid customer");
		this.require(this.customerMayAccess(customer, accountNumber), "Customer cannot access acount");
		Account account;
		account = _accounts.get(accountNumber - 1);
		int transaction = Bank.nextTransactionId();
		boolean waiting = true;
		while (waiting) {
			if (account.notLocked()) {
				try {
					account.lock(transaction);
					waiting = false;
				} catch (Error e) {
					// raced with another thread, I lost
					// continue waiting now
				}
			} else {
				try {
					Thread.sleep(1);
				} catch (InterruptedException e) {
				}
			}
		}
		tmpBalance = account.get_balance(0);
		Thread.yield();
		account.commit(transaction);
		this.invariant(this.checkSumOfAccounts(), "Inconsistencies between account balances -- checksum incorrect.");
		return tmpBalance;
	}

	/**
	 * Transfer an amount from fromAccountnr to toAccountnr
	 */
	public void transfer(Customer customer, int fromAccountnr, int toAccountnr, int amount) {
		this.invariant(this.checkSumOfAccounts(), "Inconsistencies between account balances -- checksum incorrect.");
		this.require(this.valid_customer(customer), "Invalid customer");
		this.require(this.customerMayAccess(customer, fromAccountnr), "Customer cannot access acount");
		this.require(this.customerMayAccess(customer, toAccountnr), "Customer cannot access acount");
		Account fromAccount, toAccount;
		fromAccount = _accounts.get(fromAccountnr - 1);
		toAccount = _accounts.get(toAccountnr - 1);
		int transaction = Bank.nextTransactionId();
		boolean waiting = true;
		while (waiting) {
			if (fromAccount.notLocked()) {
				try {
					fromAccount.lock(transaction);
					waiting = false;
				} catch (Error e) {
					// raced with another thread, I lost
					// continue waiting now
				}
			} else {
				try {
					Thread.sleep(1);
				} catch (InterruptedException e) {
				}
			}
		};
		waiting = true;
		while (waiting) {
			if (toAccount.notLocked()) {
				try {
					toAccount.lock(transaction);
					waiting = false;
				} catch (Error e) {
					// raced with another thread, I lost
					// continue waiting now
				}
			} else {
				try {
					Thread.sleep(1);
				} catch (InterruptedException e) {
				}
			}
		}
		fromAccount.inc_balance(-amount, 0);
		Thread.yield();
		toAccount.inc_balance(+amount, 0);
		toAccount.commit(transaction);
		fromAccount.commit(transaction);
		this.invariant(this.checkSumOfAccounts(), "Inconsistencies between account balances -- checksum incorrect.");
	}

	/**
	 * Invariant: the sum of all balances of all accounts should remain 0
	 */
	public boolean checkSumOfAccounts() {
		int tempSum = 0;

		// for (int i = 0; i < _accounts.size(); i++) {
		// tempSum = tempSum + _accounts.get(i).get_balance();
		// }

		// for (Account acc : _accounts) {
		// tempSum = tempSum + acc.get_balance();
		// }

		Iterator<Account> iterator = _accounts.iterator();
		while (iterator.hasNext()) {
			tempSum = tempSum + iterator.next().peek_balance();
		}

		return tempSum == 0;
	}

}