package internetBanking;

public abstract class Lockable extends DesignByContract {

	protected int _transactionId = 0;

	public Lockable() {
		super();
	}

	/**
	 * Answer whether no transaction is currently locking the account
	 */
	public synchronized boolean notLocked() {
		return this._transactionId == 0;
	}

	/**
	 * Answer whether this account is currently locked by the given transaction
	 * 
	 * @param transaction
	 *            identifies the transaction accessing the balance
	 */
	public synchronized boolean isLockedBy(int transaction) {
		return this._transactionId != 0;
	}

	protected abstract void clearWorkingState();

	protected abstract void commitWorkingState();

	protected abstract void copyToWorkingState();

	/**
	 * Lock this account, ensure that only transactions holding the appropriate
	 * transaction may access the account
	 * 
	 * @param transaction
	 *            identifies the transaction accessing the balance
	 */
	public synchronized void lock(int transaction) {
		this.require(this.notLocked(), "No other transaction may hold a lock.");
		this._transactionId = transaction;
		copyToWorkingState();
		this.ensure(this.isLockedBy(transaction), "Lock must be held by current transaction.");
	}

	/**
	 * Commit the pending state-changes on this account, if they were performed
	 * by the transactions holding the appropriate transaction
	 * 
	 * @param transaction
	 *            identifies the transaction accessing the balance
	 */
	public synchronized void commit(int transaction) {
		this.require(this.isLockedBy(transaction), "Lock is required to commit transaction.");
		this._transactionId = 0;
		commitWorkingState();
		this.ensure(this.notLocked(), "Lock should be released after committing the transaction.");
	}

	/**
	 * Commit the pending state-changes on this account, if they were performed
	 * by the transactions holding the appropriate transaction
	 * 
	 * @param transaction
	 *            identifies the transaction accessing the balance
	 */
	public synchronized void abort(int transaction) {
		this.require(this.isLockedBy(transaction), "Lock is required to abort transaction.");
		this._transactionId = 0;
		clearWorkingState();		
		this.ensure(this.notLocked(), "Lock should be released after aborting the transaction.");
	}

}